import pandas as pd
import numpy as np
import stan
import matplotlib.pyplot as plt
import seaborn as sb
import scipy.stats as ss
import arviz as az

pd.set_option('display.max_rows', 10000)
pd.set_option('display.max_columns', 10000)

fin_nm = input('Input file (*.xlsx) = ')
fout_nm = input('Output file (*.txt) = ')
fout = open(fout_nm, 'w')
fout.write('Input Data File = {}\n'.format(fin_nm))

xlsx_fl = pd.ExcelFile(fin_nm)
data_xlsx = pd.read_excel(xlsx_fl,  header = None)
data = data_xlsx.values
var_name = data[0]
values = data[1:]

NSt = len(data) - 1
K = len(data[0]) - 1
Sts = np.empty(NSt)
Rs = np.empty((NSt, K), dtype = 'int')
for j in range(NSt):
    Sts[j] = data[j+1][0]
    for k in range(K):
        Rs[j][k] = data[j+1][k+1]

fout.write('\nData:\n')
for j in range(NSt):
    print(f'{Sts[j]:>10g}', end = '')
    fout.write(f'{Sts[j]:>10g}')
    for k in range(K):
        print(f'{Rs[j][k]:>5d}', end = '')
        fout.write(f'{Rs[j][k]:>5d}')
    print()
    fout.write('\n')

Data = {'NSt':NSt, 'K':K, 'Sts':Sts, 'Rs':Rs}

def f_init():
    return {'a':1.0, 'b':0.1,
            'prePsis':[1.0/(NSt+1)]*(NSt+1),
            'preC':[1.0/(K-2)]*(K-2)}

with open('RatingEkman.stan', 'r') as fstan:
    sm = stan.build(fstan.read(), data = Data)

fit = sm.sample(num_chains=4, init=[f_init()]*4)

print(fit)

i_data = az.from_pystan(posterior = fit, posterior_model = sm)
smry = az.summary(i_data, var_names = ['Psis', 'sgms'])  
print('Summary =\n',smry)

az.plot_trace(i_data, var_names = ['Psis', 'sgms'])  
plt.show()

d_fm = fit.to_frame()

Psis = []
for i in range(NSt):
    Psis.append(d_fm[f'Psis.{i+1}'])
Psis = np.array(Psis).T

sgms = []
for i in range(NSt):
    sgms.append(d_fm[f'sgms.{i+1}'])
sgms = np.array(sgms).T
                
Cs = []
for i in range(K-1):
    Cs.append(d_fm[f'C.{i+1}'])
Cs = np.array(Cs).T

a = d_fm['a']   
b = d_fm['b']  

for i in range(NSt):
    sb.kdeplot(Psis.T[i], label = '$\psi_{}$'.format(i+1))
plt.title('Posterior Ditributions of $\psi_i$', fontsize = 20)
plt.yticks([])
plt.legend()
plt.tight_layout()
plt.savefig('FigPsiPost.png')
plt.show()

for i in range(NSt):
    sb.kdeplot(sgms.T[i], label = r'$\sigma_{}$'.format(i))
plt.title('Posterior Distributions of $\sigma_i$', fontsize = 20)
plt.legend()
plt.savefig('FigSgmPost.png')
plt.show()

Est_Psis = np.empty(NSt)
Est_sgms = np.empty(NSt)
for i in range(NSt):
    Est_Psis[i] = np.median(Psis.T[i])
    Est_sgms[i] = np.median(sgms.T[i])
plt.plot(Est_Psis, Est_sgms, 'o')
plt.xlabel('$\psi_i$', fontsize = 14)
plt.ylabel('$\sigma_i$', fontsize = 14)
plt.show()

plt.plot([0, 0], [0, 5], label = r'$C_0$')
for i in range(1, K-2):
    sb.kdeplot(Cs.T[i], label = r'$C_{}$'.format(i+1))
plt.plot([1,1], [0, 5], label = r'$C_{}$'.format(K-1))
plt.title('Posterior Distributions of $C_k$', fontsize = 20)
plt.yticks([])
plt.legend()
plt.savefig('FigCsPost.png')
plt.show()

med_a = np.median(a)  
med_b = np.median(b)  

v_psis = np.empty(NSt)
v_sgms = np.empty(NSt)
for i in range(NSt):
    v_psis[i] = np.median(Psis.T[i])  
    v_sgms[i] = np.median(sgms.T[i]) 

fout.write('\n{0:>15s}{1:>15s}{2:>15s}\n'.format\
           ('Stimulus', 'Med. Psi', 'Med. Sigma'))
for vst, vpsi, vsgm in zip(Sts, v_psis, v_sgms):
    fout.write('{0:>15.2f}{1:>15.2f}{2:>15.2f}\n'.format(vst, vpsi, vsgm))
    
v_Cs = np.empty(K-1)
v_Cs[0] = 0.0
v_Cs[K-2] = 1.0
for k in range(1, K-2):
    v_Cs[k] = np.median(Cs.T[k])  
for k in range(K-1):
    plt.plot([Sts[0], Sts[-1]], [v_Cs[k], v_Cs[k]], color = 'y')
plt.plot([], 'y-', label = 'C')
plt.plot(Sts, v_psis, color = 'b', lw = 3, label = '$\psi$')
plt.plot(Sts, v_psis - v_sgms, color = 'g', linestyle = '--',
         lw = 1,label = '$\psi - \sigma$')
plt.plot(Sts, v_psis + v_sgms, color = 'g', linestyle = '--',
         lw = 1, label = '$\psi+ \sigma$')
    
plt.xlabel('Stimulus', fontsize = 16)
plt.ylabel('$\psi$', fontsize = 16)
plt.legend(loc = 'upper left')
plt.title('Relation of Stimulus and $\psi$', fontsize = 18)
plt.tight_layout()
plt.savefig('PsiSgmC.png')
plt.show()

Probs = np.empty((NSt, K))
for s in range(NSt):
    Probs[s][0] = ss.norm(loc = v_psis[s], scale = v_sgms[s]).cdf(v_Cs[0])
    Probs[s][K-1] = 1.0 - ss.norm(loc = v_psis[s],
                                  scale = v_sgms[s]).cdf(v_Cs[K-2])
    for k in range(1, K-1):
        Probs[s][k] = ss.norm(loc = v_psis[s],
                              scale = v_sgms[s]).cdf(v_Cs[k]) -\
                        ss.norm(loc = v_psis[s],
                                scale = v_sgms[s]).cdf(v_Cs[k-1])
            
print('Predicted Prob(Rating|Stimulus):')
for s in range(NSt):
    for k in range(K):
        print(f'  {Probs[s][k]:7.3f}', end = '')
    print()
    
Pred_Rs = np.empty((NSt, K))
N_Rs = np.sum(Rs, axis = 1)
for i in range(NSt):
    Pred_Rs[i] = N_Rs[i] * Probs[i]
    
print('Predicted Frequencies of Rating for Stimulus:')
for s in range(NSt):
    for k in range(K):
        print(f'  {Pred_Rs[s][k]:7.1f}', end = '')
    print()

cum_data_Rs = np.cumsum(Rs, axis = 1)
cum_pred_Rs = np.cumsum(Pred_Rs, axis = 1)

print(cum_data_Rs)

for s in range(NSt):
    if s == 0:
        plt.plot(range(K), cum_data_Rs[s]/cum_data_Rs[s][K-1], 'g:',
                 lw = 3, label = 'Data')
    else:
        plt.plot(range(K), cum_data_Rs[s]/cum_data_Rs[s][K-1], 'g:',
                 lw = 3 )
    if s == 0:
        plt.plot(range(K), cum_pred_Rs[s]/cum_pred_Rs[s][K-1], 'b-',
                 label = 'Model')
    else:
        plt.plot(range(K), cum_pred_Rs[s]/cum_pred_Rs[s][K-1], 'b-')

plt.xticks(range(K), range(1, K+1))
plt.xlabel('Rating', fontsize = 14)
plt.ylabel('Cum. Prop.', fontsize = 14)
plt.legend()
plt.title('Data and Model Prediction', fontsize = 18)
plt.tight_layout()
plt.savefig('FigDataModel.png')
plt.show()

fout.close()
print('\n' + fout_nm + ' was saved.\n')
